package java.util.concurrent;

import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

public class CyclicBarrier
{
    private final ReentrantLock lock;
    private final Condition trip;
    private final int parties;
    private final Runnable barrierCommand;
    private CyclicBarrier.Generation generation;
    private int count;
    
    private void nextGeneration() {
        this.trip.signalAll();
        this.count = this.parties;
        this.generation = new CyclicBarrier.Generation();
    }
    
    private void breakBarrier() {
        this.generation.broken = true;
        this.count = this.parties;
        this.trip.signalAll();
    }
    
    private int dowait(boolean b, long awaitNanos) throws InterruptedException, BrokenBarrierException, TimeoutException {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            CyclicBarrier.Generation generation = this.generation;
            if (generation.broken) {
                throw new BrokenBarrierException();
            }
            if (Thread.interrupted()) {
                this.breakBarrier();
                throw new InterruptedException();
            }
            int count = this.count - 1;
            this.count = count;
            int n = count;
            if (n == 0) {
                boolean n2 = false;
                try {
                    Runnable barrierCommand = this.barrierCommand;
                    if (barrierCommand != null) {
                        barrierCommand.run();
                    }
                    n2 = true;
                    this.nextGeneration();
                    int n3 = 0;
                    if (n2 == 0) {
                        this.breakBarrier();
                    }
                    return n3;
                }
                finally {
                    if (n2 == 0) {
                        this.breakBarrier();
                    }
                }
            }
            while (true) {
                try {
                    if (!b) {
                        this.trip.await();
                    }
                    else if (awaitNanos > 0L) {
                        awaitNanos = this.trip.awaitNanos(awaitNanos);
                    }
                }
                catch (InterruptedException ex) {
                    if (generation == this.generation && !generation.broken) {
                        this.breakBarrier();
                        throw ex;
                    }
                    Thread.currentThread().interrupt();
                }
                if (generation.broken) {
                    throw new BrokenBarrierException();
                }
                if (generation != this.generation) {
                    return n;
                }
                if (!b || awaitNanos > 0L) {
                    continue;
                }
                this.breakBarrier();
                throw new TimeoutException();
            }
        }
        finally {
            lock.unlock();
        }
    }
    
    public CyclicBarrier(int n, Runnable barrierCommand) {
        this.lock = new ReentrantLock();
        this.trip = this.lock.newCondition();
        this.generation = new CyclicBarrier.Generation();
        if (n <= 0) {
            throw new IllegalArgumentException();
        }
        this.parties = n;
        this.count = n;
        this.barrierCommand = barrierCommand;
        return;
    }
    
    public CyclicBarrier(int n) {
        this(n, null);
    }
    
    public int getParties() {
        return this.parties;
    }
    
    public int await() throws InterruptedException, BrokenBarrierException {
        try {
            return this.dowait(false, 0L);
        }
        catch (TimeoutException cause) {
            throw new Error(cause);
        }
    }
    
    public int await(long n, TimeUnit timeUnit) throws InterruptedException, BrokenBarrierException, TimeoutException {
        return this.dowait(true, timeUnit.toNanos(n));
    }
    
    public boolean isBroken() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.generation.broken;
        }
        finally {
            lock.unlock();
        }
    }
    
    public void reset() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            this.breakBarrier();
            this.nextGeneration();
        }
        finally {
            lock.unlock();
        }
    }
    
    public int getNumberWaiting() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.parties - this.count;
        }
        finally {
            lock.unlock();
        }
    }
    
    private static class Generation
    {
        boolean broken;
        
        private Generation() {
            this.broken = false;
        }
    }
}
